import datetime


def pprint_nested_tree(tree, indent=0):
    lines = []

    if indent == 0:
        lines.append(".")

    if isinstance(tree, str):
        return tree

    # This is a bit of a hard-coded fudge for a common case that causes it
    # to be printed in a slightly more useful way.
    if tree.keys() == {"latest", "prod", "stage"}:
        entries = [
            ("latest", tree["latest"]),
            ("stage", tree["stage"]),
            ("prod", tree["prod"])
        ]
    else:
        entries = sorted(tree.items())

    for i, (key, nested_tree) in enumerate(entries, start=1):
        if i == len(entries):
            lines.append("└── " + key)

            if isinstance(nested_tree, str):
                lines[-1] = lines[-1].ljust(40) + nested_tree
            else:
                lines.extend([
                    "    " + line
                    for line in pprint_nested_tree(nested_tree, indent=indent + 1)
                ])
        else:
            lines.append("├── " + key)
            if isinstance(nested_tree, str):
                lines[-1] = lines[-1].ljust(40) + nested_tree
            else:
                lines.extend([
                    "│   " + line
                    for line in pprint_nested_tree(nested_tree, indent=indent + 1)
                ])

    return lines


def build_tree_from_paths(paths):
    tree = {}

    for path, value in paths.items():
        curr_tree = tree
        path_components = path.strip("/").split("/")
        for component in path_components:
            try:
                curr_tree = curr_tree[component]
            except KeyError:
                if component == path_components[-1]:
                    curr_tree[component] = value
                else:
                    curr_tree[component] = {}
                curr_tree = curr_tree[component]

    return tree


def pprint_path_keyval_dict(paths):
    tree = build_tree_from_paths(paths)
    return pprint_nested_tree(tree)


def pprint_date(date_obj, *, now=datetime.datetime.now()):
    assert date_obj <= now

    difference = (now - date_obj)

    is_today = date_obj.date() == now.date()
    is_yesterday = date_obj.date() == (now - datetime.timedelta(days=1)).date()

    if difference.total_seconds() <= 120:
        return "just now"
    elif is_today and difference.total_seconds() <= 60 * 60:
        return date_obj.strftime("today @ %H:%M") + " (%d min ago)" % (difference.total_seconds() // 60)
    elif is_today:
        return date_obj.strftime("today @ %H:%M")
    elif is_yesterday and difference.total_seconds() <= 60 * 60:
        return date_obj.strftime("yesterday @ %H:%M") + " (%d min ago)" % (difference.total_seconds() // 60)
    elif is_yesterday:
        return date_obj.strftime("yesterday @ %H:%M")
    else:
        result = date_obj.strftime("%a %d %B %Y @ %H:%M")

        # Trim the trailing zero from the day of the month, but not from the
        # timestamp.  Add an extra space to the month names and timestamps
        # all line up.
        if result[4] == "0":
            return result.replace(" 0", "  ", 1)
        else:
            return result



import datetime
from urllib.parse import urlparse
import uuid

import yaml

from .ecr import Ecr
from .ecs import Ecs

from .releases_store import DynamoDbReleaseStore
from .parameter_store import SsmParameterStore
from .iam import Iam

DEFAULT_ECR_NAMESPACE = "uk.ac.wellcome"


def _is_url(label):
    try:
        res = urlparse(label)
        return all([res.scheme, res.netloc])
    except ValueError:
        return False


def _load(filepath):
    with open(filepath) as infile:
        return yaml.safe_load(infile)


class Projects:
    def __init__(self, project_filepath):
        self.projects = _load(project_filepath)

    def list(self):
        return list(self.projects.keys())

    def load(self, project_id, region_name=None, role_arn=None, account_id=None):
        config = self.projects.get(project_id)

        if not config:
            raise RuntimeError(f"No matching project {project_id} in {self.projects()}")

        return Project(project_id, config, region_name, role_arn, account_id)


class Project:
    def __init__(self, project_id, config, region_name=None, role_arn=None, account_id=None):
        self.id = project_id
        self.config = config

        self.config['id'] = project_id

        if role_arn:
            self.config['role_arn'] = role_arn

        if region_name:
            self.config['aws_region_name'] = region_name

        iam = Iam(
            self.config['role_arn'],
            self.config['aws_region_name']
        )

        self.user_details = {
            'caller_identity': iam.caller_identity(),
            'underlying_caller_identity': iam.caller_identity(underlying=True)
        }

        if account_id:
            self.config['account_id'] = account_id
        else:
            if 'account_id' not in self.config:
                self.config['account_id'] = self.user_details['caller_identity']['account_id']

        # Initialise project level vars
        self.role_arn = self.config['role_arn']
        self.region_name = self.config['aws_region_name']
        self.account_id = self.config['account_id']

        # Create required services
        self.releases_store = DynamoDbReleaseStore(
            project_id=self.id,
            region_name=self.region_name,
            role_arn=self.role_arn
        )

        self.parameter_store = SsmParameterStore(
            project_id=self.id,
            region_name=self.region_name,
            role_arn=self.role_arn
        )

        self.ecr = Ecr(
            account_id=self.account_id,
            region_name=self.region_name,
            role_arn=self.role_arn
        )

        self.ecs = Ecs(
            account_id=self.account_id,
            region_name=self.region_name,
            role_arn=self.role_arn
        )

        # Ensure release store is available
        self.releases_store.initialise()
        self.prepared_releases = {}

    def _create_deployment(self, environment_id, details, description):
        return {
            "environment": environment_id,
            "date_created": datetime.datetime.utcnow().isoformat(),
            "requested_by": self.user_details['underlying_caller_identity']['arn'],
            "description": description,
            "details": details
        }

    def _create_release(self, description, images):
        release_id = str(uuid.uuid4())

        return {
            "release_id": release_id,
            "project_id": self.id,
            "project_name": self.config.get('name', 'unnamed'),
            "date_created": datetime.datetime.utcnow().isoformat(),
            "requested_by": self.user_details['underlying_caller_identity']['arn'],
            "description": description,
            "images": images,
            "deployments": []
        }

    def get_environment(self, environment_id):
        environments = {
            e['id']: e for e in self.config.get('environments', []) if 'id' in e
        }

        if environment_id not in environments:
            raise ValueError(f"Unknown environment. Expected '{environment_id}' in {environments}")

        return environments[environment_id]

    def get_deployments(self, release_id=None):
        if release_id:
            return [self.releases_store.get_release(release_id)]
        else:
            return self.releases_store.get_recent_deployments()

    def get_release(self, release_id):
        if release_id == "latest":
            return self.releases_store.get_latest_release()
        else:
            return self.releases_store.get_release(release_id)

    def get_ecs_services(self, release_id, environment_id):
        release = self.get_release(release_id)

        matched_services = {}
        for image_id, image_uri in release['images'].items():
            image_repositories = self.config.get('image_repositories')

            # Naively assume service name matches image id
            service_ids = [image_id]

            # Attempt to match deployment image id to config and override service_ids
            if image_repositories:
                matched_image_ids = [image for image in image_repositories if image['id'] == image_id]

                if matched_image_ids:
                    matched_image_id = matched_image_ids[0]
                    service_ids = matched_image_id.get('services', [])

            # Attempt to match service ids to ECS services
            available_services = [self.ecs.get_service(service_id, environment_id) for service_id in service_ids]
            available_services = [service for service in available_services if service]

            if available_services:
                matched_services[image_id] = available_services

        return matched_services

    def publish(self, namespace, image_id, label):
        self.ecr.login()

        remote_uri, remote_tag, local_tag = self.ecr.publish_image(
            namespace=namespace,
            image_id=image_id,
        )

        tag_result = self.ecr.tag_image(
            namespace=namespace,
            image_id=image_id,
            tag=remote_tag,
            new_tag=label
        )

        ssm_result = self.parameter_store.update_ssm(
            image_id,
            label,
            remote_uri,
        )

        return {
            'ecr_push': {
                'local_tag': local_tag,
                'remote_tag': remote_tag,
                'remote_uri': remote_uri,
            },
            'ecr_tag': tag_result,
            'ssm_update': ssm_result
        }

    def get_images(self, from_label, namespace=DEFAULT_ECR_NAMESPACE):
        image_repositories = self.config.get('image_repositories')

        release_images = {}
        if image_repositories:
            for image in image_repositories:
                image_id = image['id']
                account_id = image.get('account_id', self.account_id)
                namespace = image.get('namespace', namespace)

                image_details = self.ecr.describe_image(
                    namespace=namespace,
                    image_id=image_id,
                    tag=from_label,
                    account_id=account_id
                )

                release_images[image_details['image_id']] = image_details['ref']
        else:
            release_images = self.parameter_store.get_services_to_images(
                label=from_label
            )

        return release_images

    def prepare(self, from_label, description, namespace=DEFAULT_ECR_NAMESPACE):
        release_images = self.get_images(
            from_label=from_label,
            namespace=namespace
        )

        if not release_images:
            raise RuntimeError(f"No images found for {self.id}/{from_label}")

        previous_release = self.releases_store.get_latest_release()

        new_release = self._create_release(
            description=description,
            images=release_images
        )
        self.releases_store.put_release(new_release)

        return {"previous_release": previous_release, "new_release": new_release}

    def deploy(self, release_id, environment_id, namespace, description):
        release = self.get_release(release_id)
        matched_services = self.get_ecs_services(release_id, environment_id)

        # Force check for valid environment
        _ = self.get_environment(environment_id)

        deployment_details = {}

        ecs_services_deployed = {}

        # Memoize service deployments to prevent multiple deployments
        def _deploy_ecs_service(service):
            if service['serviceArn'] in ecs_services_deployed:
                return ecs_services_deployed[service['serviceArn']]
            else:
                result = self.ecs.redeploy_service(
                    service['clusterArn'],
                    service['serviceArn']
                )

                ecs_services_deployed[service['serviceArn']] = result

                return result

        for image_id, image_name in release['images'].items():
            ssm_result = self.parameter_store.update_ssm(
                service_id=image_id,
                label=environment_id,
                image_name=image_name
            )

            old_tag = image_name.split(":")[-1]
            new_tag = f"env.{environment_id}"

            tag_result = self.ecr.tag_image(
                namespace=namespace,
                image_id=image_id,
                tag=old_tag,
                new_tag=new_tag
            )

            ecs_deployments = []
            if image_id in matched_services:

                deployments = [_deploy_ecs_service(service) for service in matched_services.get(image_id)]

                for deployment in deployments:
                    service_arn = deployment['service_arn']
                    deployment_id = deployment['deployment_id']
                    ecs_deployments.append({
                        'service_arn': service_arn,
                        'deployment_id': deployment_id
                    })

            deployment_details[image_id] = {
                'ssm_result': ssm_result,
                'tag_result': tag_result,
                'ecs_deployments': ecs_deployments
            }

        deployment = self._create_deployment(environment_id, deployment_details, description)

        self.releases_store.add_deployment(release['release_id'], deployment)

        return deployment



from base64 import b64decode
import os

from botocore.exceptions import ClientError

from .iam import Iam
from .commands import cmd


class Ecr:
    def __init__(self, account_id, region_name, role_arn):
        self.account_id = account_id
        self.region_name = region_name
        self.session = Iam.get_session(
            session_name="ReleaseToolEcr",
            role_arn=role_arn,
            region_name=region_name
        )
        self.ecr = self.session.client('ecr')

        self.ecr_base_uri = (
            f"{self.account_id}.dkr.ecr.{self.region_name}.amazonaws.com"
        )

    @staticmethod
    def _get_release_image_tag(image_id):
        repo_root = cmd("git", "rev-parse", "--show-toplevel")
        release_file = os.path.join(repo_root, ".releases", image_id)

        return open(release_file).read().strip()

    @staticmethod
    def _get_repository_name(namespace, image_id):
        return f"{namespace}/{image_id}"

    def _get_full_repository_uri(self, namespace, image_id, tag):
        return f"{self.ecr_base_uri}/{Ecr._get_repository_name(namespace, image_id)}:{tag}"

    def publish_image(self, namespace, image_id):
        local_image_tag = Ecr._get_release_image_tag(image_id)
        local_image_name = f"{image_id}:{local_image_tag}"

        remote_image_tag = f"ref.{local_image_tag}"
        remote_image_name = self._get_full_repository_uri(namespace, image_id, remote_image_tag)

        try:
            cmd('docker', 'tag', local_image_name, remote_image_name)
            cmd('docker', 'push', remote_image_name)

        finally:
            cmd('docker', 'rmi', remote_image_name)

        return remote_image_name, remote_image_tag, local_image_tag

    def describe_image(self, namespace, image_id, tag, account_id=None):
        repository_name = Ecr._get_repository_name(namespace, image_id)
        if not account_id:
            account_id = self.account_id

        result = self.ecr.describe_images(
            registryId=account_id,
            repositoryName=repository_name,
            imageIds=[
                {"imageTag": tag}
            ]
        )

        if len(result['imageDetails']) == 0:
            raise RuntimeError(f"No matching images found for {repository_name}:{tag}!")

        if len(result['imageDetails']) > 1:
            raise RuntimeError(f"Multiple matching images found for {repository_name}:{tag}!")

        image_details = result['imageDetails'][0]

        is_latest = 'latest' in image_details['imageTags']

        ref_tags = [image for image in image_details['imageTags'] if image.startswith("ref.")]
        env_tags = [image for image in image_details['imageTags'] if image.startswith("env.")]

        envs = {env_tag.split('.')[-1]: self._get_full_repository_uri(
            namespace,
            image_id,
            env_tag
        ) for env_tag in env_tags}

        refs = [self._get_full_repository_uri(namespace, image_id, ref_tag) for ref_tag in ref_tags]

        # It is possible multiple ref tags can occur if images are published at new git refs with
        # no image changes, deal with it gracefully - just get the first
        if len(refs) < 1:
            raise RuntimeError(f"No matching ref tags found for {repository_name}:{tag}!")
        ref = refs[0]

        return {
            'registry_id': image_details['registryId'],
            'repository_name': image_details['repositoryName'],
            'image_digest': image_details['imageDigest'],
            'is_latest': is_latest,
            'image_id': image_id,
            'envs': envs,
            'ref': ref
        }

    def tag_image(self, namespace, image_id, tag, new_tag):
        repository_name = Ecr._get_repository_name(namespace, image_id)

        result = self.ecr.batch_get_image(
            registryId=self.account_id,
            repositoryName=repository_name,
            imageIds=[
                {"imageTag": tag}
            ]
        )

        if len(result["images"]) == 0:
            raise RuntimeError(f"No matching images found for {repository_name}:{tag}!")

        if len(result["images"]) > 1:
            raise RuntimeError(f"Multiple matching images found for {repository_name}:{tag}!")

        image = result["images"][0]

        tag_operation = {
            'source': f"{repository_name}:{tag}",
            'target': f"{repository_name}:{new_tag}"
        }

        try:
            self.ecr.put_image(
                registryId=self.account_id,
                repositoryName=repository_name,
                imageTag=new_tag,
                imageManifest=image['imageManifest']
            )

            tag_operation_status = "success"
        except ClientError as e:
            # Matching tag & digest already exists (nothing to do)
            if not e.response['Error']['Code'] == 'ImageAlreadyExistsException':
                raise e
            else:
                tag_operation_status = "noop"

        tag_operation['status'] = tag_operation_status

        return tag_operation

    def login(self):
        response = self.ecr.get_authorization_token(
            registryIds=[self.account_id]
        )

        for auth in response['authorizationData']:
            auth_token = b64decode(auth['authorizationToken']).decode()
            username, password = auth_token.split(':')
            command = ['docker', 'login', '-u', username, '-p', password, auth['proxyEndpoint']]

            cmd(*command)



import boto3


class Iam:
    def __init__(self, role_arn, region_name):
        self.session = Iam.get_session(
            session_name="ReleaseToolIamUserDetails",
            role_arn=role_arn,
            region_name=region_name
        )
        self.iam = self.session.resource('iam')

    @staticmethod
    def get_session(session_name, role_arn, region_name):
        client = boto3.client('sts')
        response = client.assume_role(
            RoleArn=role_arn,
            RoleSessionName=session_name
        )

        return boto3.session.Session(
            aws_access_key_id=response['Credentials']['AccessKeyId'],
            aws_secret_access_key=response['Credentials']['SecretAccessKey'],
            aws_session_token=response['Credentials']['SessionToken'],
            region_name=region_name
        )

    def caller_identity(self, underlying=False):
        if underlying:
            client = boto3.client('sts')
        else:
            client = self.session.client('sts')

        caller_identity = client.get_caller_identity()

        return {
            "arn": caller_identity['Arn'],
            "user_id": caller_identity['UserId'],
            "account_id": caller_identity['Account']
        }



import itertools

from .iam import Iam


class Ecs:
    def __init__(self, account_id, region_name, role_arn):
        self.account_id = account_id
        self.region_name = region_name
        self.session = Iam.get_session(
            session_name="ReleaseToolEcs",
            role_arn=role_arn,
            region_name=region_name
        )
        self.ecs = self.session.client('ecs')
        self.described_services = []
        self._load_described_services()

    def _load_described_services(self):
        service_arns_iterator = self._iterate_all_service()

        services = {}
        for cluster_arn, service_arn in service_arns_iterator:
            if cluster_arn in services:
                services[cluster_arn].append(service_arn)
            else:
                services[cluster_arn] = [service_arn]

        for cluster_arn, service_arns in services.items():
            for service_arns_chunk in Ecs._chunked_iterable(service_arns, 10):
                response = self.ecs.describe_services(
                    cluster=cluster_arn,
                    services=service_arns_chunk,
                    include=[
                        'TAGS',
                    ]
                )

                for service_details in response['services']:
                    self.described_services.append(service_details)

    # Credit to https://alexwlchan.net/2018/12/iterating-in-fixed-size-chunks/
    @staticmethod
    def _chunked_iterable(iterable, size):
        it = iter(iterable)
        while True:
            chunk = tuple(itertools.islice(it, size))
            if not chunk:
                break
            yield chunk

    def _iterate_all_service(self):
        cluster_paginator = self.ecs.get_paginator('list_clusters')
        cluster_iterator = cluster_paginator.paginate()
        service_paginator = self.ecs.get_paginator('list_services')

        for cluster_response in cluster_iterator:
            for cluster_arn in cluster_response['clusterArns']:
                service_iterator = service_paginator.paginate(
                    cluster=cluster_arn
                )

                for service_response in service_iterator:
                    for service_arn in service_response['serviceArns']:
                        yield cluster_arn, service_arn

    def redeploy_service(self, cluster_arn, service_arn):
        response = self.ecs.update_service(
            cluster=cluster_arn,
            service=service_arn,
            forceNewDeployment=True
        )

        return {
            'cluster_arn': response['service']['clusterArn'],
            'service_arn': response['service']['serviceArn'],
            'deployment_id': response['service']['deployments'][0]['id']
        }

    def get_service(self, service_id, env):
        def _match_deployment_tags(service, desired_service_tag, desired_env_tag):
            tags = service.get('tags')

            if not tags:
                return False

            env_tags = [tag for tag in tags if tag.get('key') == 'deployment:env']
            service_tags = [tag for tag in tags if tag.get('key') == 'deployment:service']

            env_tag = None
            if env_tags:
                env_tag = env_tags[0]['value']

            service_tag = None
            if service_tags:
                service_tag = service_tags[0]['value']

            if env_tag == desired_env_tag and service_tag == desired_service_tag:
                return True
            else:
                return False

        matched_services = [service for service in self.described_services if _match_deployment_tags(
            service, service_id, env
        )]

        if len(matched_services) > 1:
            raise RuntimeError(f"Multiple matching services found for {service_id}/{env}!")

        if len(matched_services) == 0:
            return None
        else:
            return matched_services[0]



from botocore.exceptions import ClientError
from boto3.dynamodb.conditions import Key

from .iam import Iam


class DynamoDbReleaseStore:
    def __init__(self, project_id, region_name, role_arn):
        self.project_id = project_id
        self.table_name = f"wellcome-releases-{project_id}"
        self.session = Iam.get_session(
            session_name="ReleaseToolDynamoDbReleaseStore",
            role_arn=role_arn,
            region_name=region_name
        )
        self.dynamo_db = self.session.resource("dynamodb")
        self.table = self.dynamo_db.Table(self.table_name)

    def describe_initialisation(self):
        return f"Create DynamoDb table {self.table_name}"

    def initialise(self):
        try:
            self.table.table_status
        except ClientError as client_error:
            if client_error.response['Error']['Code'] == 'ResourceNotFoundException':
                self._create_table()
            else:
                raise (
                    f"Unknown exception occurred while querying for {self.table_name} {client_error.response}"
                )

    def put_release(self, release):
        self.table.put_item(Item=release)

        return release

    def get_latest_release(self):
        items = self.table.query(IndexName='project_gsi',
                                 KeyConditionExpression=Key('project_id').eq(self.project_id),
                                 ScanIndexForward=False,
                                 Limit=1)
        if items['Count'] == 1:
            return items['Items'][0]
        else:
            raise ValueError(f"Latest release returned {items['Count']} results for {self.project_id}")

    def get_latest_releases(self, limit=1):
        items = self.table.query(IndexName='project_gsi',
                                 KeyConditionExpression=Key('project_id').eq(self.project_id),
                                 ScanIndexForward=False,
                                 Limit=limit)
        return items['Items']

    def get_release(self, release_id):
        items = self.table.query(KeyConditionExpression=Key('release_id').eq(release_id), Limit=1)
        if items['Count'] == 1:
            return items['Items'][0]
        else:
            raise ValueError(f"Release returned {items['Count']} results for {release_id}")

    def get_recent_deployments(self, limit=10):
        items = self.table.query(IndexName='deployment_gsi',
                                 KeyConditionExpression=Key('project_id').eq(self.project_id),
                                 ScanIndexForward=False,
                                 Limit=limit)
        return items['Items']

    def add_deployment(self, release_id, deployment, dry_run=False):
        if not dry_run:
            self.table.update_item(
                Key={
                    'release_id': release_id
                },
                UpdateExpression="SET #deployments = list_append(#deployments, :d)",
                ExpressionAttributeNames={
                    '#deployments': 'deployments',
                },
                ExpressionAttributeValues={
                    ':d': [deployment],
                }
            )

            self.table.update_item(
                Key={
                    'release_id': release_id
                },
                UpdateExpression="SET last_date_deployed = :d",
                ExpressionAttributeValues={
                    ':d': deployment['date_created'],
                }
            )

        return release_id

    def _create_table(self):
        self.dynamo_db.create_table(
            TableName=self.table_name,
            KeySchema=[
                {
                    'AttributeName': 'release_id',
                    'KeyType': 'HASH'
                }
            ],
            AttributeDefinitions=[
                {
                    'AttributeName': 'release_id',
                    'AttributeType': 'S'
                },
                {
                    'AttributeName': 'project_id',
                    'AttributeType': 'S'
                },
                {
                    'AttributeName': 'date_created',
                    'AttributeType': 'S'
                },
                {
                    'AttributeName': 'last_date_deployed',
                    'AttributeType': 'S'
                }
            ],
            GlobalSecondaryIndexes=[
                {
                    'IndexName': 'project_gsi',
                    'KeySchema': [
                        {
                            'AttributeName': 'project_id',
                            'KeyType': 'HASH'
                        },
                        {
                            'AttributeName': 'date_created',
                            'KeyType': 'RANGE'
                        },
                    ],
                    'Projection': {
                        'ProjectionType': 'ALL'
                    },
                    'ProvisionedThroughput': {
                        'ReadCapacityUnits': 1,
                        'WriteCapacityUnits': 1
                    }
                },
                {
                    'IndexName': 'deployment_gsi',
                    'KeySchema': [
                        {
                            'AttributeName': 'project_id',
                            'KeyType': 'HASH'
                        },
                        {
                            'AttributeName': 'last_date_deployed',
                            'KeyType': 'RANGE'
                        },
                    ],
                    'Projection': {
                        'ProjectionType': 'ALL'
                    },
                    'ProvisionedThroughput': {
                        'ReadCapacityUnits': 1,
                        'WriteCapacityUnits': 1
                    }
                }
            ],
            ProvisionedThroughput={
                'ReadCapacityUnits': 1,
                'WriteCapacityUnits': 1
            }
        )



import click
import json

from dateutil.parser import parse
from pprint import pprint
from tabulate import tabulate

from . import git
from .pretty_printing import pprint_date
from .project import Projects

DEFAULT_PROJECT_FILEPATH = ".wellcome_project"
DEFAULT_ECR_NAMESPACE = "uk.ac.wellcome"


def _format_ecr_uri(uri):
    image_name = uri.split("/")[2]
    image_label, image_tag = image_name.split(":")

    return {
        'label': image_label,
        'tag': image_tag
    }


def _summarise_ssm_response(images):
    for image in images:
        yield {
            'name': image['Name'],
            'value': image['Value'],
            'last_modified': image['LastModifiedDate'].strftime('%d-%m-%YT%H:%M')
        }


def _summarise_release_deployments(releases):
    for release in releases:
        for deployment in release['deployments']:
            try:
                requested_by = deployment.get("requested_by", {}).get("id", "")
            except AttributeError:
                requested_by = deployment.get("requested_by", "")

            yield {
                'release_id': release.get('release_id'),
                'environment_id': deployment.get('environment'),
                'deployed_date': parse(deployment.get('date_created')),
                'requested_by': requested_by.split('/')[-1],
                'description': deployment.get('description')
            }


@click.group()
@click.option('--project-file', '-f', default=DEFAULT_PROJECT_FILEPATH)
@click.option('--verbose', '-v', is_flag=True, help="Print verbose messages.")
@click.option('--confirm', '-y', is_flag=True, help="Non-interactive deployment confirmation")
@click.option("--project-id", '-i', help="Specify the project ID")
@click.option("--region-name", '-i', help="Specify the AWS region name")
@click.option("--account-id", help="Specify the AWS account ID")
@click.option("--role-arn", help="Specify an AWS role to assume")
@click.option('--dry-run', '-d', is_flag=True, help="Don't make changes.")
@click.pass_context
def cli(ctx, project_file, verbose, confirm, project_id, region_name, account_id, role_arn, dry_run):
    try:
        projects = Projects(project_file)
    except FileNotFoundError:
        if ctx.invoked_subcommand != "initialise":
            message = f"Couldn't find project metadata file {project_file!r}.  Run `initialise`."
            raise click.UsageError(message) from None
        ctx.obj = {
            'project_filepath': project_file,
            'verbose': verbose,
            'dry_run': dry_run,
        }
        return

    project_names = projects.list()
    project_count = len(project_names)

    if not project_id:
        if project_count == 1:
            project_id = project_names[0]
        else:
            project_id = click.prompt(
                text="Enter the project ID",
                type=click.Choice(project_names)
            )

    project = projects.load(
        project_id=project_id,
        region_name=region_name,
        account_id=account_id,
        role_arn=role_arn
    )

    config = project.config

    user_arn = project.user_details['caller_identity']['arn']
    underlying_user_arn = project.user_details['underlying_caller_identity']['arn']

    if verbose:
        click.echo(click.style(f"Loaded {project_file}:", fg="cyan"))
        pprint(config)
        click.echo("")
        click.echo(click.style(f"Using role_arn:         {config['role_arn']}", fg="cyan"))
        click.echo(click.style(f"Using aws_region_name:  {config['aws_region_name']}", fg="cyan"))
        click.echo(click.style(f"Running as role:        {user_arn}", fg="cyan"))
        if user_arn != underlying_user_arn:
            click.echo(click.style(f"Underlying role:        {underlying_user_arn}", fg="cyan"))

        click.echo(click.style(f"Using account_id:       {config['account_id']}", fg="cyan"))
        click.echo("")

    ctx.obj = {
        'project_filepath': project_file,
        'verbose': verbose,
        'confirm': confirm,
        'dry_run': dry_run,
        'project': project
    }


def _publish(project, image_id, namespace, label):
    click.echo(click.style(f"Attempting to publish {project.id}/{image_id}", fg="green"))

    publish_result = project.publish(
        namespace=namespace,
        image_id=image_id,
        label=label
    )

    ssm_path = publish_result['ssm_update']['ssm_path']
    ssm_value = publish_result['ssm_update']['image_name']
    local_tag = publish_result['ecr_push']['local_tag']
    remote_uri = publish_result['ecr_push']['remote_uri']
    tag_source = publish_result['ecr_tag']['source']
    tag_target = publish_result['ecr_tag']['target']

    click.echo(click.style(f"Published {local_tag} to {remote_uri}", fg="yellow"))
    click.echo(click.style(f"Tagged {tag_source} with {tag_target}", fg="yellow"))
    click.echo(click.style(f"Updated SSM path {ssm_path} to {ssm_value}", fg="yellow"))

    click.echo(click.style(f"Done publishing {project.id}/{image_id}", fg="bright_green"))


@cli.command()
@click.option("--image-id", required=True)
@click.option("--namespace", required=True, default=DEFAULT_ECR_NAMESPACE)
@click.option("--label", default="latest")
@click.pass_context
def publish(ctx, image_id, namespace, label):
    project = ctx.obj['project']

    _publish(
        project=project,
        image_id=image_id,
        namespace=namespace,
        label=label
    )


def _deploy(project, release_id, environment_id, namespace, description, confirm=True):
    release = project.get_release(release_id)

    environment = project.get_environment(environment_id)
    env_id = environment.get('id')
    env_name = environment.get('name', environment_id)

    click.echo("")
    click.echo(click.style(f"Deploying release {release['release_id']}", fg="green"))
    click.echo(click.style(f"Targeting env: {env_id}, ({env_name})", fg="yellow"))
    click.echo(click.style(f"Requested by: {release['requested_by']}", fg="yellow"))
    click.echo(click.style(f"Date created: {release['date_created']}", fg="yellow"))

    ecs_services = project.get_ecs_services(release_id, environment_id)

    rows = []

    headers = ["image ID", "services"]

    for image_id, services in sorted(ecs_services.items()):
        service_names = [serv["serviceArn"].split("/")[-1] for serv in services]
        rows.append([image_id, ", ".join(sorted(service_names))])

    print("ECS services discovered:\n")
    print(tabulate(rows, headers=headers))

    if not confirm:
        click.echo("")
        click.confirm(click.style("Create deployment?", fg="cyan", bold=True), abort=True)

    result = project.deploy(release_id, environment_id, namespace, description)

    click.echo("")
    click.echo(click.style("Deployment Summary", fg="green"))
    click.echo(click.style(f"Requested by: {result['requested_by']}", fg="yellow"))
    click.echo(click.style(f"Date created: {result['date_created']}", fg="yellow"))
    click.echo("")
    for image_id, summary in result['details'].items():
        click.echo(click.style(f"Summary for {image_id}", fg="bright_yellow"))
        click.echo(click.style(
            f"{image_id}: SSM Updated {summary['ssm_result']['ssm_path']} to {summary['ssm_result']['image_name']}",
            fg="yellow"
        ))
        click.echo(click.style(
            f"{image_id}: ECR Updated {summary['tag_result']['source']} to {summary['tag_result']['target']}",
            fg="yellow"
        ))

        for ecs_deployment in summary['ecs_deployments']:
            click.echo(click.style(
                f"{image_id}: ECS Deployed {ecs_deployment['service_arn']} to {ecs_deployment['deployment_id']}",
                fg="yellow"
            ))

    click.echo("")
    click.echo(click.style(f"Deployed release {release['release_id']} to {env_id}, ({env_name})", fg="bright_green"))


@cli.command()
@click.option('--release-id', prompt="Release ID to deploy", default="latest", show_default=True,
              help="The ID of the release to be deployed, or the latest release if unspecified")
@click.option('--environment-id', prompt="Environment ID to deploy release to",
              default="stage", show_default=True,
              help="The target environment of this deployment")
@click.option("--namespace", default=DEFAULT_ECR_NAMESPACE, show_default=True)
@click.option('--description', prompt="Enter a description for this deployment",
              help="A description of this deployment", default="No description provided")
@click.pass_context
def deploy(ctx, release_id, environment_id, namespace, description):
    confirm = ctx.obj['confirm']
    project = ctx.obj['project']

    _deploy(
        project=project,
        release_id=release_id,
        environment_id=environment_id,
        namespace=namespace,
        description=description,
        confirm=confirm
    )


def _prepare(project, from_label, namespace, description):
    release = project.prepare(
        from_label=from_label,
        namespace=namespace,
        description=description
    )

    prev_release = release["previous_release"]
    new_release = release["new_release"]

    click.echo(click.style(f"Prepared release from images in {from_label}", fg="green"))
    click.echo(click.style(f"Requested by: {new_release['requested_by']}", fg="yellow"))
    click.echo(click.style(f"Date created: {new_release['date_created']}", fg="yellow"))
    click.echo("")

    rows = []

    headers = [
        "service",
        "old image",
        "new image",
        "Git commit",
    ]

    for service, image in sorted(new_release["images"].items()):
        # The image IDs are of the form
        #
        #     {ecr_repo_uri}/{namespace}/{service}:ref.{git_commit}
        #
        prev_git_commit = prev_release["images"].get(service, "").split(".")[-1][:7]
        new_git_commit = image.split(".")[-1][:7]

        rows.append([
            service,
            prev_git_commit,
            new_git_commit
            if new_git_commit == prev_git_commit
            else click.style(new_git_commit, fg="green"),
            git.log(new_git_commit),
        ])

    print(tabulate(rows, headers=headers))

    click.echo("")
    click.echo(click.style(f"Created release {new_release['release_id']}", fg="bright_green"))

    return new_release['release_id']


@cli.command()
@click.option('--from-label', prompt="Label to base release upon",
              help="The existing label upon which this release will be based",
              default="latest", show_default=True)
@click.option('--namespace', help="Namespace in which to locate images",
              default=DEFAULT_ECR_NAMESPACE)
@click.option('--description', prompt="Description for this release",
              default="No description provided")
@click.pass_context
def prepare(ctx, from_label, namespace, description):
    project = ctx.obj['project']

    _prepare(
        project=project,
        from_label=from_label,
        namespace=namespace,
        description=description
    )


@cli.command()
@click.option('--from-label', prompt="Label to base release upon",
              help="The existing label upon which this release will be based", default="latest", show_default=True)
@click.option('--environment-id', prompt="Environment ID to deploy release to",
              default="stage", show_default=True,
              help="The target environment of this deployment")
@click.option("--namespace", default=DEFAULT_ECR_NAMESPACE, show_default=True)
@click.option('--description', prompt="Enter a description for this deployment",
              help="A description of this deployment", default="No description provided")
@click.pass_context
def release_deploy(ctx, from_label, environment_id, namespace, description):
    project = ctx.obj['project']
    confirm = ctx.obj['confirm']

    release_id = _prepare(
        project=project,
        from_label=from_label,
        namespace=namespace,
        description=description
    )

    _deploy(
        project=project,
        release_id=release_id,
        environment_id=environment_id,
        namespace=namespace,
        description=description,
        confirm=confirm
    )


@cli.command()
@click.argument('release_id', required=False)
@click.pass_context
def show_release(ctx, release_id):
    project = ctx.obj['project']

    if not release_id:
        release_id = 'latest'

    release = project.get_release(release_id)

    click.echo(json.dumps(release, sort_keys=True, indent=2))


@cli.command()
@click.argument('release_id', required=False)
@click.pass_context
def show_deployments(ctx, release_id):
    project = ctx.obj['project']

    releases = project.get_deployments(release_id)

    rows = []

    headers = [
        "release ID",
        "environment ID",
        "deployed date",
        "request by",
        "description"
    ]

    for summary in _summarise_release_deployments(releases):
        try:
            environment_id = summary["environment_id"].get("id", "")
        except AttributeError:
            environment_id = summary["environment_id"]

        rows.append([
            summary["release_id"],
            environment_id,
            pprint_date(summary["deployed_date"]),
            summary["requested_by"],
            summary["description"]
        ])

    print(tabulate(rows, headers=headers))


@cli.command()
@click.option('--label', '-l', help="The label to show (e.g., latest')")
@click.pass_context
def show_images(ctx, label):
    project = ctx.obj['project']

    if not label:
        label = 'latest'

    images = project.get_images(label)

    rows = []

    headers = [
        "image ID",
        "label",
        "tag",
    ]

    for image_id, ecr_uri in images.items():
        ecr_uri = _format_ecr_uri(ecr_uri)
        rows.append([image_id, ecr_uri["label"], ecr_uri["tag"]])

    print(tabulate(rows, headers=headers))


def main():
    cli()
